
/* 
 * Copyright 2010  Simon Pickartz, Chair for Operating Systems,
 *                                 RWTH Aachen University
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "RCCE.h"
#include "iRCCE.h"



int RCCE_APP(int argc, char **argv)
{
	int num_ranks;
	int root_rank, my_rank; 
	iRCCE_SEND_REQUEST send_request;
	int buffer;

	/* initialize the library */
	RCCE_init(&argc, &argv);
	iRCCE_init();

	my_rank   = RCCE_ue();
	num_ranks = RCCE_num_ues();
	root_rank = 0;

	/* check parameters */
	if( num_ranks < 3 )
	{
		if(my_rank == 0)
			fprintf( stderr, 
				"wheel of fortune needs at least 3 UEs; "
				"try again\n");
		exit(-1);
	}

	/* master's part */
	if( my_rank == root_rank ) {
		int i;
		int buf[RCCE_MAXNP];
		iRCCE_RECV_REQUEST recv_request[RCCE_MAXNP];

		printf( "\n"
				"#####################################\n"
				"## Welcome to wheel of fortune:    ##\n"
				"##                                 ##\n"
				"## Number of players: %d            ##\n"
				"## Rank of root: %d                 ##\n"
				"#####################################\n\n",
				num_ranks, my_rank );

		/* wait for the others to guess the number */
		for( i=1; i<num_ranks; ++i ) {
			iRCCE_irecv( (char*)&buf[i], 4, iRCCE_ANY_SOURCE, &recv_request[i] );
		}
		iRCCE_irecv_wait( NULL );

		int my_num = (double)rand()*100/RAND_MAX;
		int winner = 1;
		int min = 100;

		printf( "My number\t\t: %d\n", my_num );

		/* determine the winner of the game */
		for( i=1; i<num_ranks; ++i ) {
			printf( "Number of rank %d\t: %d\n", i, buf[i] );
			if( ((my_num-buf[i])*(my_num-buf[i])) < min ) {
				min = ((my_num-buf[i])*(my_num-buf[i]));
				winner = i;
			}
		}

		printf( "The winner is: %d!\n", winner );
	}
	/* players' part */
	else {
		/* guess a number */
		srand( my_rank );
		int buf = (double)rand()*100/RAND_MAX;
		iRCCE_SEND_REQUEST send_request;

		/* send result to master */
		iRCCE_isend( (char*)&buf, 4, root_rank, &send_request );	

		iRCCE_isend_wait( &send_request );
	}


	RCCE_finalize();

	return 0;
}

