
#include "RCCE.h"
#include "iRCCE.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

int RCCE_APP(int argc, char **argv)
{
  int t, Tmax = 1000;

  int i, I, j, J;
  int my_rank;
  int num_ranks;

  int n, N = 200;
  int m, M = 200;

  double** NewValues;
  double** OldValues;

  iRCCE_SEND_REQUEST upper_send_req;
  iRCCE_SEND_REQUEST lower_send_req;

  iRCCE_RECV_REQUEST upper_recv_req;
  iRCCE_RECV_REQUEST lower_recv_req;

  iRCCE_WAIT_LIST waitlist;

  RCCE_init(&argc, &argv);
  iRCCE_init();
  iRCCE_init_wait_list(&waitlist);

  my_rank   = RCCE_ue();
  num_ranks = RCCE_num_ues();

  m = M;
  J = 0;

  n = N/num_ranks;
  if(my_rank == num_ranks-1) n += N%num_ranks;
  I = n*my_rank;

  printf("(%d) %d x %d / offsets: %d, %d / (%d x %d)\n", my_rank, N, M, I, J, n, m);


#ifdef _LAPLACE_SHMEM_

  NewValues = (double**)malloc((N+2)*sizeof(double*));    
  NewValues[0] = (double *)RCCE_shmalloc((N+2)*(M+2)*sizeof(double));  

  OldValues = (double**)malloc((N+2)*sizeof(double*));    
  OldValues[0] = (double *)RCCE_shmalloc((N+2)*(M+2)*sizeof(double));  

  for (i = 1; i < N+2; i++)
  {
    NewValues[i] = NewValues[i-1] + (M+2);
    OldValues[i] = OldValues[i-1] + (M+2);
  }

  RCCE_barrier(&RCCE_COMM_WORLD);

#else

  NewValues = (double**)malloc((n+2)*sizeof(double*));    
  NewValues[0] = (double *)malloc((n+2)*(m+2)*sizeof(double));

  OldValues = (double**)malloc((n+2)*sizeof(double*));    
  OldValues[0] = (double *)malloc((n+2)*(m+2)*sizeof(double));

  for (i = 1; i < n+2; i++)
  {
    NewValues[i] = NewValues[i-1] + (m+2);
    OldValues[i] = OldValues[i-1] + (m+2);
  }

#endif

  printf("(%d) Memory allocated!\n", my_rank); fflush(stdout);
       
#ifdef _LAPLACE_SHMEM_
  
  if(my_rank == 0)
  {
    for(i=0; i<N+2; i++)
    {
      for(j=0; j<M+2; j++)
      {	
	NewValues[i][j] = 0.0;
	if (i==0)  NewValues[i][j] = 100.0;
	if(i==N+1) NewValues[i][j] = 100.0;
      }
      NewValues[i][0]   = 100.0;
      NewValues[i][N+1] = 100.0;
    }

    for(i=0; i<N+2; i++)
    {
      for(j=0; j<M+2; j++)
      {
	OldValues[i][j] = NewValues[i][j];
      }
    }
  }
  RCCE_barrier(&RCCE_COMM_WORLD);

#else

  for(i=0; i<n+2; i++)
  {
    for(j=0; j<m+2; j++)
    {	
      NewValues[i][j] = 0.0;
    }
  }
  
  if(my_rank == 0)
  {
    for(j=0; j<m+2; j++)
    {
      NewValues[0][j] = 100.0;
    }
  }

  if(my_rank == num_ranks-1)
  {
    for(j=0; j<m+2; j++)
    {
      NewValues[n+1][j] = 100.0;
    }
  }
  
  for(i=0; i<n+2; i++)
  {
    NewValues[i][0]   = 100.0;
    NewValues[i][N+1] = 100.0;
  }
  
  for(i=0; i<n+2; i++)
  {
    for(j=0; j<m+2; j++)
    {
      OldValues[i][j] = NewValues[i][j];    
    }
  }

  RCCE_barrier(&RCCE_COMM_WORLD);

#endif

  printf("(%d) Arrays initialized!\n", my_rank); fflush(stdout);

  for(t=0; t<Tmax; t++)
  {
    for(i=1; i<n+1; i++)
    {
      for(j=1; j<m+1; j++)
      {
#ifdef _LAPLACE_SHMEM_
	NewValues[I+i][J+j] = 0.25 * ( OldValues[I+i-1][J+j] + OldValues[I+i+1][J+j] +
	                               OldValues[I+i][J+j-1] + OldValues[I+i][J+j+1] );
#else
	NewValues[i][j] = 0.25 * ( OldValues[i-1][j] + OldValues[i+1][j] +
	                           OldValues[i][j-1] + OldValues[i][j+1] );
#endif
      }
    }

#ifdef _LAPLACE_SHMEM_

    RCCE_barrier(&RCCE_COMM_WORLD);
    
#else

    if(my_rank != 0)
    {
      iRCCE_isend((char*)NewValues[1], (m+2)*sizeof(double), my_rank-1, &upper_send_req);
      iRCCE_irecv((char*)OldValues[0], (m+2)*sizeof(double), my_rank-1, &upper_recv_req);
      iRCCE_add_to_wait_list(&waitlist, &upper_send_req, &upper_recv_req);
    }

    if(my_rank != num_ranks-1)
    {
      iRCCE_irecv((char*)OldValues[n+1], (m+2)*sizeof(double), my_rank+1, &lower_recv_req);
      iRCCE_isend((char*)NewValues[n],   (m+2)*sizeof(double), my_rank+1, &lower_send_req);
      iRCCE_add_to_wait_list(&waitlist, &lower_send_req, &lower_recv_req);
    }

    iRCCE_wait_all(&waitlist);   
    
#endif

#ifdef _LAPLACE_SHMEM_

    for(i=1; i<n+1; i++)
    {
      for(j=1; j<m+1; j++)
      {
	OldValues[I+i][J+j]=NewValues[I+i][J+j];
      }
    }
    RCCE_barrier(&RCCE_COMM_WORLD);

#else

    for(i=1; i<n+1; i++)
    {
      for(j=1; j<m+1; j++)
      {
	OldValues[i][j]=NewValues[i][j];
      }
    }

#endif
  }

  printf("(%d) Algorithm completed!\n", my_rank); fflush(stdout);

  RCCE_barrier(&RCCE_COMM_WORLD);

#if 0

#ifdef _LAPLACE_SHMEM_
  if(my_rank == 0)
  {
    for(i=1; i<N+1; i++)
    {
      for(j=1; j<M+1; j++)
      {
	printf("%3.2f ", NewValues[i][j]);
      }
      printf("\n");
    }
  }
#else
  
  sleep(my_rank);

  for(i=1; i<n+1; i++)
  {
    for(j=1; j<m+1; j++)
    {
      printf("%3.2f ", NewValues[i][j]);
    }
    printf("\n");
  }
#endif

#endif
  
  RCCE_finalize();

  return 0;
}
